import argparse, sys, re
from os import path
from oobfuzz import OOBFuzz

payloads = path.join(path.dirname(__file__), "data", "payloads")

parser = argparse.ArgumentParser(prog='OOBFuzz', epilog=f'Examples of payloads can be found at: {payloads}/')
parser.add_argument('--payloads', type=str, help='Path to directory containing payloads', required=True)
parser.add_argument('--output', type=str, help='File to output result to')
parser.add_argument('--threads', type=int, help='Number of threads to fuzz (default: 5)')
parser.add_argument('--target', type=str, help='Single target to run against')
parser.add_argument('--targets', type=str, help='File with newline seperator including several targets')
parser.add_argument('--exclude', type=str, help='String of response codes seperated by "," (example 404,401,403) to exclude from printing')
parser.add_argument('--proxy', type=str, help='Proxy to route through (example "http://localhost:8080")')
parser.add_argument('--blocks', type=int, help='Amount of blocks registered to exit (default: 5). This means that if 5 blocks are registered the fuzzer will exit')
parser.add_argument('--redir', help='Allow HTTP redirects', action="store_true")
parser.add_argument('--urls', type=str, help='File with newline seperator including endpoints. This will skip fetch through Gau')
parser.add_argument('--stdin-targets', nargs=argparse.REMAINDER)
parser.add_argument('stdin', nargs='?', type=argparse.FileType('r'), default=sys.stdin)



if len(sys.argv) == 1:
    parser.print_help(sys.stderr)
    sys.exit(1)

args = parser.parse_args()

if not sys.stdin.isatty():
    stdin_targets = parser.parse_args().stdin.read().splitlines()
else:
    stdin_targets = None


if args.proxy:
    proxyMatch = re.findall(r'(http(?:s)?://[a-zA-Z0-9\.-]+:[\d]{1,5})', args.proxy)
    if len(proxyMatch) <= 0:
        parser.print_help(sys.stderr)
        exit(1)

'''if args.help_me:
    help_me()
    sys.exit(0)'''

payloads_dir = args.payloads
output = args.output if args.output else None
threads = args.threads if args.threads else 5
target = args.target if args.target else None
targets = args.targets if args.targets else None
exclude = args.exclude if args.exclude else None
proxy = args.proxy if args.proxy else None
blocks = args.blocks if args.blocks else 5
redir = args.redir if args.redir else False
urls = args.urls if args.urls else None

oobfuzz = OOBFuzz(output, threads, target, targets, exclude, proxy, blocks, redir, urls, stdin_targets, payloads_dir)


#!/usr/bin/python3
import requests
import sys, concurrent.futures, re
from datetime import datetime
from subprocess import PIPE, Popen
from os import devnull, walk, path
from urllib.parse import quote
from urllib3 import disable_warnings
from random import choice
from time import sleep
from socket import gaierror
from OpenSSL import SSL
from base64 import b64encode

class OOBFuzz():

    def __init__(self, output, threads, target, targets, exclude, proxy, blocks, redir, urls, stdin_targets, payloads_dir):

        disable_warnings()

        self.gau = True
        self.proxy = None
        if proxy:
            self.proxy = {"http": proxy, "https": proxy}
        self.output = output
        self.threads = threads
        self.targets = []
        self.blocks = blocks
        self.redir = redir
        if path.exists(payloads_dir):
            self.payloads_dir = payloads_dir
            if not self.payloads_dir[-1] == '/':
                self.payloads_dir + '/'
        else:
            print(f'Couldn\'t open supplied directory of payloads at: {payloads_dir}')
            sys.exit(1)

        if exclude:
            self.exclude = exclude.split(',')
        else:
            self.exclude = exclude

        self.paramRE = re.compile(r'(([\w]+)=([^&]+)+)')
        
        self.payloads = []
        self.excludedParams = ['submit']


        (_, _, filenames) = next(walk(f'{self.payloads_dir}'))
        for filename in filenames:
            with open(f'{self.payloads_dir}{filename}', 'r') as f:
                tmpDict = {filename[:filename.index(".")]: []}
                for payload in f:
                    tmpDict[filename[:filename.index(".")]].append(quote(payload.rstrip()))
            self.payloads.append(tmpDict)

        self.useragents = []
        with open(f'{path.join(path.dirname(__file__), "data", "user-agents.txt")}', 'r') as agents:
            for useragent in agents:
                self.useragents.append(useragent.rstrip())

        if targets:
            try:
                with open(f'{targets}', 'r') as targetsFile:
                    for targetName in targetsFile:
                        self.targets.append(targetName.rstrip())
            except (FileNotFoundError):
                print(f'Unable to open {targets}, ensure proper permissions or that the file exists..')
                sys.exit(1)
        elif target:
            self.targets.append(target)
        elif urls:
            self.gau = False
            try:
                with open(f'{urls}', 'r') as targetsFile:
                    for targetName in targetsFile:
                        self.targets.append(targetName.rstrip())
            except (FileNotFoundError):
                print(f'Unable to open {targets}, ensure proper permissions or that the file exists..')
                sys.exit(1)
        elif stdin_targets:
            self.targets = stdin_targets

        print("Status\tLength\tTime\tHost")
        print("---------------------------------")
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.threads) as executor:
            worker = executor.map(self.run, [target for target in self.targets])
            for _ in worker:
                pass

    def run(self, target):

        if self.gau:
            with open(devnull, 'w') as dn:
                process = Popen(['/usr/bin/gau', '-subs', target], stdout=PIPE, stderr=dn)
            stdout = process.communicate()[0].decode("utf-8").split("\n")
            urls = [url for url in stdout if '=' in url]

            urls = list(set(urls))
        else:
            urls = list(set(self.targets))

        result = []
        
        for baseurl in urls:
            matches = self.paramRE.findall(baseurl)

            for match in matches:
                if str(match[1]).lower() in self.excludedParams:
                    continue
                value = match[2]
                for d in self.payloads:
                    for _, payloadList in d.items():
                        for payload in payloadList:
                            if self.blocks == 0:
                                # We're out of luck, exit
                                break
                            user_agent = choice(self.useragents)
                            headers = {'User-Agent': user_agent}
                            payload = payload.strip('\n')
                            base = baseurl.replace(value, payload)
                            if 'ID_BASE64' in base:
                                url = base.replace('ID_BASE64', b64encode(bytes(base, encoding='utf-8')).decode('utf-8'))
                            else:
                                url = base
                            try:
                                r = requests.get(url, headers=headers, allow_redirects=self.redir, proxies=self.proxy, verify=False)
                                # Here we attempt to circumvent rate limit and/or temporary blocks on our IP
                                # Often when 403 og 429 is returned because of a block a small html document
                                # is returned with that is len =~ 200
                                if r.status_code == 403 and len(r.content) < 200 or r.status_code == 429 and len(r.content) < 200:
                                    # Testing has shown that at least blocks from Akamai last around 2 minutes
                                    print(f"{str(datetime.now())} - Request was possibly blocked by WAF: Sleeping thread for 140 seconds for {url}")
                                    sleep(140)
                                    r = requests.get(url, headers=headers, allow_redirects=self.redir, proxies=self.proxy, verify=False)
                                    if r.status_code == 403 and len(r.content) < 200 or r.status_code == 429 and len(r.content) < 200:
                                        # At this point we just continue
                                        self.blocks -= 1
                                        continue
                            except (SSL.SysCallError, gaierror):
                                continue
                            if self.exclude and str(r.status_code) in self.exclude:
                                continue
                            result.append([r.status_code, len(r.content), str(r.elapsed.total_seconds()*1000)[:7], url])
                            print(f'{r.status_code}\t{len(r.content)}\t{str(r.elapsed.total_seconds()*1000)[:7]}\t{url}')
                            if r.status_code == 404:
                                # The url/endpoint is not working, lets just break out
                                break
                            '''if self.save_responses and len(r.content) != 0:
                                try:
                                    with open(f'{self.output_dir}/{str(datetime.now()).replace(" ", "T").replace(":", "-").split(".")[0]}-{self.payloaddata.index(payload)}', 'w') as f:
                                        f.write(f'{url}\n\n{r.status_code}\n\n{r.content}')
                                except:
                                    print(f'Error: could not write file {self.output_dir}/{str(datetime.now()).replace(" ", "T").replace(":", "-").split(".")[0]}-{self.payloaddata.index(payload)}')'''
                            sleep(0.5)                
        return result


from sys import exit
from subprocess import Popen, PIPE
from os import environ
from setuptools import setup, find_packages
from setuptools.command.install import install

with open('README.md', 'r') as desc:
    long_desc = desc.read()

class InstallDependencies(install):

    def run(self):
        
        # Install Go
        try:
            # Debian based
            process = Popen(['/usr/bin/apt-get', 'install', 'golang', '-y'], stdout=PIPE)
        except IOError:
            # Fedora/Red hat 
            process = Popen(['/usr/bin/dnf', 'install', 'golang', '-y'], stdout=PIPE)
        process.communicate()[0].decode('utf-8')

        # Get GAU dependency
        process = Popen(['/usr/bin/go', 'get', '-u', 'github.com/lc/gau'], stdout=PIPE)
        process.communicate()[0].decode('utf-8')
        
        try:
            process = Popen(['/bin/cp', f'{environ["HOME"]}/go/bin/gau', '/usr/bin/gau'], stdout=PIPE)
        except EnvironmentError:
            print('This installer requires root permissions to copy Gau into /usr/bin/gau')
            exit(1)
        process.communicate()[0].decode('utf-8')

        install.run(self)

setup(
        name = 'OOB Fuzz',
        version = '0.1.0',
        author = 'Casper G. Nielsen',
        author_email = 'whopsec@protonmail.com',
        description = 'Conduct OOB Fuzzing of targets with payloads towards callback server',
        long_description = long_desc,
        long_description_content_type = 'text/markdown',
        url = 'https://github.com/CasperGN/oobfuzz',
        cmdclass={'install': InstallDependencies},
        packages = find_packages(),            
        include_package_data = True,
        install_requires=[
            'requests',
            'pyopenssl',
            'cffi',
        ],
        classifiers = [
            'Programming Language :: Python :: 3',
            'License :: OSI Approved :: MIT License',
        ],
        python_requires = '>=3.4',
)